#!/usr/bin/perl
# zim.pl 0.3.0 - A PERL/tk front end to Zephyrgrams
#
# Copyright: (C) 1999, Gabriel A. Cohen
#                      <gac3@acpub.duke.edu>
#                      http://www.duke.edu/~gac3/zim.html
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#   For revisional changes, see ChangeLog file
#
######################################################################

#everyone loves tk
use Tk; 
use Tk::ROText;

sub build_online_list; #updates contents of user listbox
sub process_stream;    #main event processing loop
sub mesg_user;         #send a zwrite
sub status_change;     #parse a status change event
sub get_message;       #parse an incoming message
sub array_ref;         #allocate a new array
sub trim;              #eliminate bad white space 
sub parse_conf;        #read in configuration file

#initialize options to default
my(%options)=();
$options{"logfile"}="/dev/null";
$options{"incomingevent"}="";
$options{"statusevent"}="";
$options{"incolor"}="blue";
$options{"outcolor"}="red";
$options{"dispbackcolor"}="grey";
$options{"inputbackcolor"}="grey";
$options{"inputforecolor"}="black";


#clear out old daemons
open PS, "ps -u $ENV{'USER'} | grep zwgc |";
while(<PS>)
{
    @line=split;
    #print "Killing $line[0] zwgc process!\n";
    kill 9,$line[0]; 
}
close PS;

#begin daemon
open ZWGC, "zwgc -ttymode -f zimrc |";

#create top level window
my($top)=new MainWindow;

#create split frames
my($topframe)=$top->Frame();
$topframe->pack(-anchor=>'n',-fill=>'both',-side=>'top',-expand=>1);
my($bottomframe)=$top->Frame();
$bottomframe->pack(-anchor=>'s',-fill=>'both',-side=>'bottom');

#stick label at top
$label=$topframe->Label(-text => 'Zephyr Instant Messenger');
$label->pack(-anchor=>'n',-fill=>'x',-side=>'top');

#build message dialog display
my($display)= $topframe->ROText(-width => 40,-height => 10);
$display->pack(-anchor=>'w',-fill=>'both',-side=>'left',-expand=>1);

#add a scroll bar
$dispscroll = $topframe->Scrollbar(-command => ['yview', $display]); 
$dispscroll->pack(-anchor=>'e',-fill=>'y',-side=>'right');

#list box for user ids
my($box) = $bottomframe->Listbox(-relief => 'sunken',-width => -1, -height=>5,-setgrid => 'yes', -selectmode=>'single');

#add scroll bar for list box
$friendscroll = $bottomframe->Scrollbar(-command => ['yview', $box]); 

$friendscroll->pack(-anchor=>'e',-fill=>'y',-side=>'right');
$box->pack(-anchor=>'e',-fill=>'y',-side=>'right');

#build the input box
my($input)= $bottomframe->Text(-width => 25,-height => 5,-background=>'white');
$input->pack(-anchor=>'w',-fill=>'both',-side=>'left',-expand=>1);

my (%status)=();

#construct hash table for buddies
open ANYONE, "$ENV{'HOME'}/.anyone";
while(<ANYONE>)
{
    @line=split;
    $name=$line[0];
    $info=array_ref();
    @{$info}=(0,"nowhere","never");
    $status{$name}=$info;
}

#update info for whoever is currently online
open ZNOL, "znol |";
while(<ZNOL>)
{
    #check date, eliminate old ones
    @line=split(":");
    @other=split;
    $pointer=$status{$line[0]};
    @{$pointer}[0]=${$pointer}[0]+1;
    @{$pointer}[1]=$other[1];
    @{$pointer}[2]=$other[6];
}

#update user list box
build_online_list;

#add callback for zwgc pipe
$top->fileevent(ZWGC, 'readable', sub{process_stream(%status,$box,$top,$display)});
#add callback for message sending
$input->bind('<KeyPress-Return>',sub{mesg_user($input,$display,$box)});

#read in the configuration
parse_conf;

#set up display options
$display->tagConfigure("outfg", -foreground => $options{"outcolor"});
$display->tagConfigure("infg", -foreground => $options{"incolor"});
$display->configure(-background => $options{"dispbackcolor"});
$input->configure(-background=>$options{"inputbackcolor"},
    -foreground=>$options{"inputforecolor"});


#init tk
MainLoop;

#the main loop for processing zwgc output
sub process_stream
{    
    #grab the zim tag
    $line=<ZWGC>;
    if($line =~ m/message/)
    {
	#zwgc incoming message	
	system($options{"incomingevent"});

        #grab info about message and message
	($name, $date, $loc, @mesg)=get_message();	

	#log the event
	open LOG, ">> $options{'logfile'}";
    	print LOG "$name on $date\n@mesg\n\n";
	close LOG;
	
        #insert into display
	foreach $element (@mesg)
	{
	    $display->insert('end',"$name -> $element\n",["infg"]);
	}
    }
    elsif($line =~ m/znol/)
    {
	#zwgc status message
	system($options{"statusevent"});	

	#grab mesg
	status_change;

	#update the listbox
	build_online_list;
    }
    #scroll to most recent text
    $display->yview(moveto=>1);
}

#send a zwrite to a user
sub mesg_user
{
    my ($input,$display,$box) = @_;
    
    #get selected user
    $dest=$box->get('active');
    
    #store message in a string
    $message=$input->get('1.0','end');
  
    
    #clear input box
    $input->delete('1.0','end');
    
    #send mesg
    $message=trim($message);    
    open MESG, "| zwrite $dest > /dev/null";
    print MESG "$message\n";
    close MESG;
  
    #add to dialog  
    $display->insert('end',"$dest <- $message\n",["outfg"]);
    
    
    #log the event
    open LOG, ">> $options{'logfile'}";
    print LOG "Message To $dest\n$message\n\n";
    close LOG;

    #scroll display
    $display->yview(moveto=>1);
}

#parse a status message, altering %status
sub status_change
{
    #read in name and event type
    $name=<ZWGC>;
    $type=<ZWGC>;
    $type=trim($type);
    $name=trim($name);
    
    #adjust user status accordingly
    $pointer=$status{$name};
    if($type=~m/in/)
    {
	@{$pointer}[0]=${$pointer}[0]+1;
    }
    else
    {
	@{$pointer}[0]=${$pointer}[0]-1;
    }

    #not used right now
    $date=<ZWGC>; 
    $date=trim($date);
    $loc=<ZWGC>;
    $loc=trim($loc);
    $end=<ZWGC>;
}

#parse an incoming message
sub get_message
{
    @mesg=();

    #read in fields
    $name=<ZWGC>;
    $date=<ZWGC>;
    $loc=<ZWGC>;
    
    #trim fields
    $name=trim($name);
    $date=trim($date);
    $loc=trim($loc);
    
    #grab body of mesg
    while(<ZWGC>)
    {
	if($_ =~ m/<\/zim>/)
	{return ($name,$date,$loc,@mesg);}
	chomp;
	push(@mesg,trim($_));
    }
}

#update whats in the user listbox
sub build_online_list
{
    #get selected user
    $dest=$box->get('active');
    
    #clean out old dudes
    $box->delete(0,'end');

    #iterate through users
    foreach $guy (sort (keys %status))
    {
	$ptr=$status{$guy};
	
	#if online, insert into box
	if(@{$ptr}[0]>0)
        {
	    $box->insert('end', $guy);
	    if($guy eq $dest)
            {
		$box->activate('end');
	    }
        }
    }
}

#create a new array and return pointer
sub array_ref
{
    my @array;
    return \@array;
}

#trim off bad whitespace
sub trim
{
    my ($string)=@_;
    $string=~s/^\s+//;
    $string=~s/\s+$//;
    return $string;
}

#reads in configuration file
sub parse_conf
{
    my(@line)=();

    #open file
    open CONF, "zim.conf";
    while(<CONF>)
    {
	@line=split(':');

	#if a valid option, update value in options map
	if(defined($options{$line[0]}))
	{
	    $line[1]=trim($line[1]);
	    #print "setting option $line[0] to ";
	    $options{$line[0]}=$line[1]; 
	    #print "$options{$line[0]}\n";
	}	
    }
    close CONF;
}
